/// \file  
/// \date 2008-10-10
/// \author Xavier Michelon, Toradex SA 
///  
/// \brief Sample program for using Oak device in Linux


#include "OakHidBase.h"
#include "OakFeatureReports.h"
#include <stdio.h>
#include <stdlib.h>
#include <string>


using namespace std;
using namespace Toradex::Oak;


// define some local constants
namespace {
   string const& kDefaultDevice("/dev/usb/hiddev1"); ///< Default device, if not specified as an argument
}

/// \brief wrapper function for error report that add file, line and function in which an error occured
#define CHECK_OAK_CALL(status) checkOakCall(status, __FILE__, __LINE__, __FUNCTION__)


//************************************************************************************************* 
/// \brief Check the status code of a Oak library call. 
///
/// If the status is not OK, this function display an error message to the standard error output
/// then quit the application
///
/// Note This function is not to be called directly, but via the CHECK_OAK_CALL macro that
/// insert the line, file and function where the error occured
///
/// \param[in] status the status code to check
/// \param[in] file The file where the error occured
/// \param[in] line The line number where the error occured
/// \param[in] function The name of the function where the error occured
//************************************************************************************************* 
void checkOakCall(EOakStatus status, const char *file, int line, const char* function)
{
   if (eOakStatusOK != status)
   {
      fprintf(stderr, "Error: %s in function %s (file %s, line %d)\n", getStatusString(status).c_str(), 
              function, file, line);
      exit(1);
   }
}


//************************************************************************************************* 
/// \brief The program entry point
///
/// \param[in] argc The number of arguments of the program
/// \param[in] argv The argument list of the program
//************************************************************************************************* 
int main(int argc, char **argv)
{
   // if the device file is not provided as an argument, we use a default value
   std::string device(argc >= 2 ? argv[1] : kDefaultDevice);
   
   int deviceHandle;
   CHECK_OAK_CALL(openDevice(device, deviceHandle));
   
   // get the device informations
   DeviceInfo devInfo;
   CHECK_OAK_CALL(getDeviceInfo(deviceHandle, devInfo));
   
   printf("Device Name: %s\n", devInfo.deviceName.c_str());
   printf("Volatile User Device Name: %s\n", devInfo.volatileUserDeviceName.c_str());
   printf("Persistent User Device Name: %s\n", devInfo.persistentUserDeviceName.c_str());
   printf("Serial Number: %s\n", devInfo.serialNumber.c_str());
   printf("VendorID: 0x%x :: ProductID: 0x%x :: Version 0x%x\n", devInfo.vendorID, devInfo.productID,
          devInfo.version);   
   printf("Number of channels: %d\n", devInfo.numberOfChannels);


   // Set the report Mode
   CHECK_OAK_CALL(setReportMode(deviceHandle, eReportModeAfterSampling, true));
   EOakReportMode reportMode;
   CHECK_OAK_CALL(getReportMode(deviceHandle, reportMode, true));
   printf("Report mode: %u\n", int(reportMode));
   
   // Set the LED Mode
   CHECK_OAK_CALL(setLedMode(deviceHandle, eLedModeOff, true));
   EOakLedMode ledMode;
   CHECK_OAK_CALL(getLedMode(deviceHandle, ledMode, true));
   printf("Led Mode: %u\n", int(ledMode));
   
   // Set the report rate
   CHECK_OAK_CALL(setReportRate(deviceHandle, 2000, true));
   unsigned int reportRate;
   CHECK_OAK_CALL(getReportRate(deviceHandle, reportRate, true));
   printf("Report rate: %u\n", reportRate);
   
   // Set the sample rate
   CHECK_OAK_CALL(setSampleRate(deviceHandle, 2000, true));
   unsigned int sampleRate;
   CHECK_OAK_CALL(getSampleRate(deviceHandle, sampleRate, true));
   printf("Sample rate: %u\n", sampleRate);
   
   // Retrieve Channel infos
   std::vector<ChannelInfo> channelInfos(devInfo.numberOfChannels);
   for (int i = 0; i < devInfo.numberOfChannels; i++)
   {
      ChannelInfo& chanInfo = channelInfos[i]; 
      CHECK_OAK_CALL(getChannelInfo(deviceHandle, i, chanInfo));
      printf("\nChannel %d:\n", i);
      printf("   Name: %s\n", chanInfo.channelName.c_str());
      printf("   Persistent user channel name: %s\n", chanInfo.persistentUserChannelName.c_str());
      printf("   Volatile user channel name: %s\n", chanInfo.volatileUserChannelName.c_str());
      printf("   Data type: %d bits (%ssigned)\n", chanInfo.bitSize, chanInfo.isSigned ? "" : "un");
      printf("   Unit exponent: %d\n", chanInfo.unitExponent);
      printf("   Unit %s (USB code 0x%x)\n", chanInfo.unit.c_str(), chanInfo.unitCode);
   }
   printf("\n");
   
   // Now we loop several times displaying what we read in interrupt reports
   for (int i = 0; i < 3; i++)
   {
      std::vector<int> values;
      CHECK_OAK_CALL(readInterruptReport(deviceHandle, values));
      for (int i = 0; i < devInfo.numberOfChannels; i++)
      {
         ChannelInfo& chanInfo = channelInfos[i];
         printf("%s: ", chanInfo.channelName.c_str());
         if (chanInfo.isSigned)
            printf("%d",values[i]);
         else
            printf("%u", values[i]);
         if (0 != chanInfo.unitExponent)
            printf("x10e%d", chanInfo.unitExponent);
         printf("\n");   
      }
      printf("\n");
   }
      
   // Cleanup
   CHECK_OAK_CALL(closeDevice(deviceHandle));
   return 0;

}



